.include "tn45def.inc"
;*******************************************************************************
;*                                                                             *
;* Programm fr einen RTC der mit DCF-77 synchronisiert ist                    *
;*                                                                             *
;* Eine mit dem internen Oszillator realisierte Uhr wird durch den Empfang     *
;* des DCF-77 Zeitsignals synchronisiert.                                      *
;* Das Auslesen der aktuellen Zeit ist ber den IIC-Bus mglich. Dabei kann    *
;* ein PCF8583 (A2h") oder DS1307 (D0h) simuliert werden.                      *
;*                                                                             *
;* Alle fr die Zeit nicht bentigte Speicherstellen knnen ber den IIC-Bus   *
;* geschrieben und gelesen werden. Beim Tiny 45 sind dies 256 Byte.            *
;*                                                                             *
;* Mit dem Tiny 45 kann der IIC-Bus mit 80KHz getaktet werden.                 *
;*                                                                             *
;* Noch zu verbessern ist:                                                     * 
;* 1. beliebiger Portpin fr DCF77_IN (begonnen)                               *
;* 2. Konfigurationsmglichkeit wie in PCF8583 bzw. DS1307 (Interrupt/Alarm)   *
;* 3. AM/PM Bit bei DS1307                                                     *
;* 4. 12 Stunden-Modus bei DS1307                                              *
;* 5. Stromspar-Mglichkeiten des Tiny ausschpfen                             *
;*                                                                             *
;* Copyright Michal Fluhr (fluhri@gmx.de)                                      *
;*                                                                             *
;*******************************************************************************

.def		tmp			= r16

; IIC_Slave
;.equ		DevIICAdr	= $A0		; PCF8583 Device-IIC-Address
;.equ		DevIICAdr	= $A2		; PCF8583 Device-IIC-Address
.equ		DevIICAdr	= $D0		; DS1307 Device-IIC-Address

.equ		IntOut		= PinB3		; Ausgang fr Alarm oder Sekundentakt


;IIC
.equ		SDA			= PinB0
.equ		SCL			= PinB1


.def		IICDat		= r18		; IIC-Bus-Daten
.def		IICAdr		= r19		; IIC-Bus-Adresse
.def		IICSub		= r30		; IIC-Sub-Speicheradresse (kann direkt als Pointer benutzt werden)
						; r31 ist auch belegt, wird fr Z bentigt
.def		tmpIIC		= r16

.def		BitCount	= r20		; das wievielte Bit wird gelesen
.def		ByteCount	= r23
.def		IICState	= r21		; Bit 0-3 => das wievielte Byte wird gelesen
									; Bit 7   => letztes Acknowledge

.def		LastPort	= r22


; DCF-77
.equ		DCF77_In	= PinB2

.def		DCF77Second	= r1
.def		DCF77Buffer	= r8		; bis r15
.def		CntOverflow	= r17		; Da mit gte Zeit von Timer 1 nur 0,5s sind, mssen die berlufe gezhlt werden
.def		RamPointer	= r28
						; r29 ist auch belegt, wird fr Y bentigt
.def		tmpDCF		= r26

; RTC
.def		Cnt0OverFlow= r27

				.cseg

				.org $0000
				rjmp RESET 					; Reset handler
.IF DCF77_In==PinB2 
				rjmp DCF77_StartSec			; IRQ0 handler
				rjmp IIC_StateChanged		; Pin change handler
.ELSE
				rjmp NoneInt				; IRQ0 handler
				rjmp PinChanged				; Pin change handler
.ENDIF
				rjmp DCF77_ReadBit			; Timer1 Compare Match A
				rjmp DCF77_StartMin			; Timer1 overflow handler
				rjmp NoneInt				; Timer0 overflow handler
				rjmp NoneInt				; EEPROM ready
				rjmp NoneInt				; Analog Comparator handler
				rjmp NoneInt				; ADC Conversion Complete
				rjmp NoneInt				; Timer1 Compare Match B
				rjmp Second150				; Timer0 Compare Match A
				rjmp NoneInt				; Timer0 Compare Match B
				rjmp NoneInt				; Watchdog Time-out
				rjmp NoneInt				; USI Start
				rjmp NoneInt				; USI Overflow


NoneInt:		reti

Reset:			rcall	InitIIC				; IIC-Pins initialisieren

				; IIC-Buffer lschen
				ldi 	tmpIIC,$00
				ldi		r30,$60
				clr		r31
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC
				inc		r30
				st		Z,tmpIIC

				ldi		tmp,$80				; Set Clock Prescale at 1
				out		CLKPR,tmp
				ldi		tmp,$00
				out		CLKPR,tmp
				
				rcall	InitDCF77
				rcall	InitTime

				clr		r31

				in		LastPort,PinB
				ldi		IICSub,$60

				ser		tmp					; Interupt-Flags zurcksetzen
				out		GIFR,tmp

				ldi		tmp,$FF
				out		TIFR,tmp			; Clear Timer Interrupt Flags

				sei							; alle Interrupts freigeben
				
Main:			

	       		rjmp	Main

;******************************************************************************
;* Wenn der DCF77-Eingang nicht PinB3 ist, mu nach einem PinChanged          *
;* entschieden werden, woher der Interrupt kommt.                             *
;******************************************************************************
.IF DCF77_In!=PinB2
PinChanged:		in		tmp,PinB
				eor		tmp,LastPort

				sbrs	tmp,DCF77_IN
				rjmp 	IIC_StateChanged	; Interrupt fr IIC-Bus
				in		LastPort,PinB
				rjmp 	DCF77_StartSec		; Interrupt vom DCF77-Empfnger
.ENDIF
;******************************************************************************
;* Vom DCF77-Signal unabhngige Uhr											  *
;******************************************************************************
InitTime:		sbi		DDRB,IntOut			; IntOut als Ausgang
				sbi		PortB,IntOut
				
				ldi		tmp,250				; 8MHz/256/250 ergeben 150 berlufe pro Sekunde
				out		OCR0A,tmp			; Set Timer 0 Compare Value A

				ldi		tmp,$04				; CK/256
				out		TCCR0B,tmp

				; Bei erreichen von OCR0A, Zhler auf 0 setzen
				clr		tmp
				ori		tmp,1<<WGM01
				out		TCCR0A,tmp

				in		tmp,TIMSK
				ori		tmp,1<<OCIE0A		; Enable Timer 0 Compare Match A Interrupt
				out		TIMSK,tmp			; Timer/Counter Interrupt Mask

				clr		Cnt0OverFlow

				ret

Second150:		; mu 150 Mal pro Sekunde aufgerufen werden
				inc		Cnt0OverFlow
				cpi		Cnt0OverFlow,63
				brsh	Second150_0	
				sbi		PortB,IntOut
				rjmp	Second150End

Second150_0:	cbi		PortB,IntOut
				cpi		Cnt0OverFlow,125
				brlo	Second150End
				
				clr		Cnt0OverFlow
			.IF (DevIICAdr==$A0)|(DevIICAdr==$A2)
				ldi		RamPointer,$62
			.ELIF DevIICAdr==$D0
				ldi		RamPointer,$60
			.ENDIF
				clr		r29

				ld		tmp,Y				; Sekunde ist um	
				inc		tmp
				rcall	BCDKorrektur
				st		Y,tmp

				cpi		tmp,$60
				brlo	Second150End		; Minute ist noch nicht um
				clr		tmp					; Minute ist um => Sekunden auf 0 setzen
				st		Y+,tmp
				ld		tmp,Y				; Minute incrementieren
				inc		tmp
				rcall	BCDKorrektur
				st		Y,tmp

				cpi		tmp,$60
				brlo	Second150End		; Stunde ist noch nicht um
				clr		tmp					; Stunde ist um => Minuten auf 0 setzen
				st		Y+,tmp
				ld		tmp,y				; Stunde incrmentieren
				inc		tmp
				rcall	BCDKorrektur
				st		Y,tmp

				cpi		tmp,$24
				brlo	Second150End		; Tag ist noch nicht um
				clr		tmp					; Tag ist um => Stunden auf 0 setzen
				st		Y+,tmp
				
				ld		tmp,Y				; Wochentag incrementieren
			.IF (DevIICAdr==$A0)|(DevIICAdr==$A2)
				inc		tmp					; Wochentag von 0-6
				cpi		tmp,7
				brlo	Second150_1
				clr		tmp
Second150_1:	
			.ELIF DevIICAdr==$D0
				cpi		tmp,$07				; Wochentag von 1-7
				brlo	Second150_2
				clr		tmp
Second150_2:	inc		tmp
			.ENDIF
				st		Y+,tmp

				ld		tmp,Y				; Tag incrementieren
				inc		tmp
				rcall	BCDKorrektur
				st		Y,tmp

				rcall	CalcTagMonat

				brlo	Second150End		; Monat ist noch nicht um
				ld		tmp,Y				; Monat ist um => Tag auf 1 setzen
				ldi		tmp,1
				st		Y+,tmp
				ld		tmp,Y				; Monat incrementieren
				inc		tmp
				rcall	BCDKorrektur
				st		Y,tmp

				cpi		tmp,$13
				brlo	Second150End		; Jahr ist noch nicht um
				ld		tmp,Y				; Jahr ist um => Monat auf 1 setzen
				ldi		tmp,$01
				st		Y+,tmp
				ld		tmp,Y				; Jahr incrementieren
				inc		tmp
				rcall	BCDKorrektur
				st		Y+,tmp
				
Second150End:	reti


BCDKorrektur:	push	tmp
				andi	tmp,$0F
				cpi		tmp,$0A
				brlo	BCDKorrekturNo
				
				pop		tmp
				subi	tmp,$0A
				swap	tmp
				inc		tmp
				swap	tmp
				push	tmp
				andi	tmp,$F0
				cpi		tmp,$A0
				brlo	BCDKorrekturNo
				subi	tmp,$A0
				rjmp	BCDKorrekturEnd

BCDKorrekturNo: pop		tmp

BCDKorrekturEnd:ret

;******************************************************************************
;* Berechnet zu dem Monat und dem Jahr die Anzahl der Tage                    *
;*                                                                            *
;* Input:  Y steht auf Tag     A = Monat; R0 = Jahr                           *
;* Output: tmp = Anzahl Tage                                                  *
;* Use:                                                                       *
;* Stack:                                                                     *
;******************************************************************************
CalcTagMonat:	ld		tmp,Y+				; Monat holen
				ld		tmp,Y+

				cpi		tmp,2
				brne	CalcTMNoFebruar

CalcFebruar:	ld		tmp,Y
				andi	tmp,$03
				cpi		tmp,0
				brne	Calc28

Calc29:			ldi		tmp,$29
				rjmp	CalcEnde

Calc28:			ldi		tmp,$28
				rjmp	CalcEnde

CalcTMNoFebruar:ror		tmp
                brcs	CalcUngerade

CalcGerade:		subi	tmp,8
				brcs	Calc30

Calc31:			ldi		tmp,$31
				rjmp	CalcEnde

CalcUngerade:	subi	tmp,$08
				brcs	Calc31

Calc30:			ldi		tmp,$30

CalcEnde:		sbiw	RamPointer,2
				ret		

;******************************************************************************
;* DCF-77                                                                     *
;*                                                                            *
;* DCF77_In mu an Int0 angeschlossen werden                                  *
;*                                                                            *
;******************************************************************************
InitDCF77:		cbi		DDRB,DCF77_In		; DCF77_In als Eingang
;				sbi		PortB,DCF77_In		; mit PullUp

				in		tmpDCF,MCUCR
				clr		tmpDCF
				ori		tmpDCF,1<<ISC00		; The rising edge of INT0 generates an interrupt request
				ori		tmpDCF,1<<ISC01
				out		MCUCR,tmpDCF

.IF DCF77_In==PinB2 
				in		tmpDCF,GIMSK
				ori		tmpDCF,1<<INT0		; External Interrupt Request 0 Enable
				out		GIMSK,tmpDCF
.ELSE
				in		tmp,PCMSK
				ori 	tmp,1<<DCF77_IN
				out		PCMSK,tmp			; Only DCF77_IN Pin Change Interrrupt
.ENDIF

				ldi		tmpDCF,$0F			; CK/16384
				out		TCCR1,tmpDCF

				ldi		tmpDCF,$48			; entspricht 156 ms
				out		OCR1A,tmpDCF		; Set Timer 1 Compare Value A
				
				in		tmpDCF,TIMSK
				ori		tmpDCF,1<<OCIE1A	; Enable Timer 1 Compare Match A Interrupt
				ori		tmpDCF,1<<TOIE1		; Enable Timer 1 Overflow Interrupt
				out		TIMSK,tmpDCF		; Timer/Counter Interrupt Mask

				clr		tmpDCF
				mov		DCF77Second,tmpDCF
			.IF (DevIICAdr==$A0)|(DevIICAdr==$A2)
				sts		$0062,DCF77Second
			.ELIF DevIICAdr==$D0
				sts		$0060,DCF77Second
			.ENDIF
				mov		CntOverflow,tmpDCF

				mov		r31,DCF77Buffer		; Clear Buffer
				st		Z,tmpDCF
				inc		r31
				st		Z,tmpDCF
				inc		r31
				st		Z,tmpDCF
				inc		r31
				st		Z,tmpDCF
				inc		r31
				st		Z,tmpDCF
				inc		r31
				st		Z,tmpDCF
				inc		r31
				st		Z,tmpDCF
				inc		r31
				st		Z,tmpDCF

				ret

;******************************************************************************
;* Routinen zum Speichern der gltigen DCF77-Zeit							  *
;******************************************************************************
.Macro			PCF8583				; legt Uhrzeit und Datum PCF8583 entsprechend ab
				ldi		RAMPointer,$63
				clr		r29

				; Minute
				mov		tmpDCF,r11
				ror		tmpDCF
				mov		tmpDCF,r12
				ror		tmpDCF
				ror		tmpDCF
				andi	tmpDCF,$7F
				st		Y+,tmpDCF

				; Stunde
				mov		tmpDCF,r11
				ror		tmpDCF
				ror		tmpDCF
				andi	tmpDCF,$3F
				st		Y+,tmpDCF

				; Jahr/Tag
				mov		tmpDCF,r10
				ror		tmpDCF
				andi	tmpDCF,$3F
				bst		r8,7
				bld		tmpDCF,6
				bst		r8,0
				bld		tmpDCF,7
				st		Y+,tmpDCF

				; Wochentag/Monat
				mov		tmpDCF,r9
				mov		CntOverflow,r10
				rol		CntOverflow
				rol		tmpDCF			; MMMM MWWW

				swap	tmpDCF			; MWWW MMMM
				rol		tmpDCF			; WWWM MMM0 M
				lsr		tmpDCF			; 0WWW MMMM M
				rol		tmpDCF			; WWWM MMMM
				
				subi	tmpDCF,$20
				st		Y+,tmpDCF
.ENDMACRO

.MACRO			DS1307
				ldi		RAMPointer,$61
				clr		r29
				
				; Minute
				mov		tmpDCF,r11
				ror		tmpDCF
				mov		tmpDCF,r12
				ror		tmpDCF
				ror		tmpDCF
				andi	tmpDCF,$7F
				st		Y+,tmpDCF

				; Stunde
				mov		tmpDCF,r11
				ror		tmpDCF
				ror		tmpDCF
				andi	tmpDCF,$3F
;				AM/PM
				st		Y+,tmpDCF

				; Wochentag
				mov		tmpDCF,r10
				rol		tmpDCF
				mov		tmpDCF,r9
				rol		tmpDCF
				andi	tmpDCF,$07
				st		Y+,tmpDCF

				; Tag
				mov		tmpDCF,r10
				ror		tmpDCF
				andi	tmpDCF,$3F
				st		Y+,tmpDCF

				; Monat/Jahrhundert
				mov		tmpDCF,r9
				ror		tmpDCF
				ror		tmpDCF
				andi	tmpDCF,$1F
				st		Y+,tmpDCF

				; Jahr
				mov		tmpDCF,r9
				rol		tmpDCF
				mov		tmpDCF,r8
				rol		tmpDCF
				st		Y+,tmpDCF
				
.ENDMACRO


; Aufruf bei einer steigender Flanke von DCF_77
DCF77_StartSec:	cpi		CntOverflow,3
				brlo	DCF77_IncSec
				clr		DCF77Second
			.IF (DevIICAdr==$A0)|(DevIICAdr==$A2)
				sts		$0062,DCF77Second
			.ELIF DevIICAdr==$D0
				sts		$0060,DCF77Second
			.ENDIF
				rjmp	DCF77_InitInt

DCF77_IncSec:	inc		DCF77Second
				mov		tmp,DCF77Second
				andi	tmp,$0F
				cpi		tmp,$0A
				brlo	DCF77_WriteSec
				; Sekunden als BCD-Zahl inkrementieren
				; Zehner-Sekunden hochzhlen und Einer-Sekunden lschen 
				ldi		tmp,$10
				add		tmp,DCF77Second
				andi	tmp,$F0
				mov		DCF77Second,tmp

DCF77_WriteSec:	
;			.IF (DevIICAdr==$A0)|(DevIICAdr==$A2)
;				sts		$0062,DCF77Second
;			.ELIF DevIICAdr==$D0
;				sts		$0060,DCF77Second
;			.ENDIF

DCF77_InitInt:	clr		tmpDCF
				out		TCNT1,tmpDCF

				in		tmpDCF,TIFR
				ori		tmpDCF,1<<OCF1A		; Clear Interrupt Flag
				out		TIFR,tmpDCF

				in		tmpDCF,TIMSK
				ori		tmpDCF,1<<OCIE1A	; Enable Timer 1 Compare Match A Interrupt
				out		TIMSK,tmpDCF		; Timer/Counter Interrupt Mask
				
				clr		CntOverflow

				reti


; Aufruf durch Timer Compare A nach 150 ms
DCF77_ReadBit:	;sbi		PortB,IntOut
				in		tmpDCF,TIMSK
				andi	tmpDCF,$BF			; Disable Timer 1 Compare Match A Interrupt
				out		TIMSK,tmpDCF		; Timer/Counter Interrupt Mask

				clc
				sbic	PinB,DCF77_In		; Bit lesen
				sec

				ror		DCF77Buffer
				ror		r9
				ror		r10
				ror		r11
				ror		r12
				ror		r13
				ror		r14
				ror		r15

; Ergebnis
;r8        r9        r10       r11       r12       r13       r14       r15
;PJJJJJJJ  JMMMMMWW  WTTTTTTP  SSSSSSPM  MMMMMMSA  ZZAR____  ________  ___
;38421842  11842142  12184212  21842114  218421 2  211
; 0000      0         00       00     0  00 

DCF77_Test:		;cpi		Second,59
				;brne	DCF77_ReadEnd


DCF77_ReadEnd:	reti


; Aufruf durch Timer OverFlow nach 0,5s
DCF77_StartMin:	inc		CntOverflow
				
				cpi		CntOverflow,2
				brne	DCF77_StartMin1
				
				inc		DCF77Second
;			.IF (DevIICAdr==$A0)|(DevIICAdr==$A2)
;				sts		$0062,DCF77Second
;			.ELIF DevIICAdr==$D0
;				sts		$0060,DCF77Second
;			.ENDIF
				
;				sbi		PortB,IntOut
								
				rjmp	DCF77_StartMinEnd

DCF77_StartMinEndZ:		rjmp	DCF77_StartMinEnd
DCF77_BreakZ:	rjmp	DCF77_Break

DCF77_StartMin1:;cbi		PortB,IntOut
				cpi		CntOverflow,3
				brne	DCF77_StartMinEndZ	; erst nach drei berlufen sind 1,5s vorbei
				
				; berprfung der empfangen Bits
				ldi		tmp,38
				cp		DCF77Second,tmp
				brlo	DCF77_StartMinEndZ	; Es wurden zu wenige Bits empfangen

				; Paritt der Minute prfen
				mov		CntOverflow,r11
				ror		CntOverflow
				mov		tmp,r12
				ror		tmp
				ror		CntOverflow
				ror		tmp

				ldi		CntOverflow,$01
				ldi		RamPointer,$00
DCF77_Paritaet3:ror		tmp
				brcc	DCF77_Paritaet4
				eor		tmp,CntOverflow
DCF77_Paritaet4:inc		RamPointer
				cpi		RamPointer,$07
				brne	DCF77_Paritaet3

				ror		tmp
				brcs	DCF77_BreakZ

				; Paritt der Stunde prfen
				mov		tmp,r10
				ror		tmp
				mov		tmp,r11
				ror		tmp
				lsr		tmp

				ldi		RamPointer,$00
DCF77_Paritaet1:ror		tmp
				brcc	DCF77_Paritaet2
				eor		tmp,CntOverflow
DCF77_Paritaet2:inc		RamPointer
				cpi		RamPointer,$07
				brne	DCF77_Paritaet1

				ror		tmp
				brcs	DCF77_Break

				; Paritt von Tag, Wochentag, Monat und Jahr berechnen
				mov		tmp,r10
				andi	tmp,$F7	
				
				ldi		RamPointer,$00
DCF77_Paritaet5:ror		tmp
DCF77_Paritaet6:brcc	DCF77_Paritaet7
				eor		tmp,CntOverflow
DCF77_Paritaet7:inc		RamPointer
				cpi		RamPointer,$07
				brlo	DCF77_Paritaet5

				brne	DCF77_Paritaet8
				; Nchstes Register R9 laden
				ror		tmp
				mov		tmp,r9
				rjmp	DCF77_Paritaet6

DCF77_Paritaet8:cpi		RamPointer,$0F
				brlo	DCF77_Paritaet5

				brne	DCF77_Paritaet9
				; Nchstes Register R8 laden
				ror		tmp
				mov		tmp,r8
				rjmp	DCF77_Paritaet6

DCF77_Paritaet9:cpi		RamPointer,$17
				brlo	DCF77_Paritaet5

				ror		tmp
				brcs	DCF77_Break

				

			.IF (DevIICAdr==$A0)|(DevIICAdr==$A2)
		    	PCF8583
			.ELIF DevIICAdr==$D0
		      	DS1307
			.ENDIF

DCF77_Break:	ldi		CntOverflow,$03

DCF77_StartMinEnd:	
				reti	

;******************************************************************************
;* IIC-Slave                                                                  *
;*                                                                            *               
;* SDA und SCL mssen als Pin von PortB definiert sein                        *               
;*                                                                            *               
;* Der PinChange-Interrupt mu Function IIC_StateChanged aufrufen             *               
;*                                                                            *               
;*                                                                            *               
;******************************************************************************

InitIIC:
				cbi		DDRB,SDA			; SDA als Eingang
				sbi		PortB,SDA			; mit PullUp
				cbi		DDRB,SCL			; SCL als Eingang
				sbi		PortB,SCL			; mit PullUp

				in		tmp,GIMSK
				ori		tmp,1<<PCIE			; Pin Change Interrupt Enable
				out		GIMSK,tmp		

				in		tmp,PCMSK
				ori		tmp,1<<SCL
				ori 	tmp,1<<SDA
				out		PCMSK,tmp			; Only SCL and SDA Pin Change Interrrupt

				ldi		IICSub,$60

				ret


IIC_StateChanged:
PinChangeEnd:	reti

